#include <bsnmp/snmpmod.h>

#include <sys/param.h>
#include <sys/resource.h>
#include <sys/queue.h>

#include <errno.h>
#include <kvm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>

#include "cpuload_tree.h"

static struct lmodule *module;
static long int       cp_time[CPUSTATES], cp_load[CPUSTATES];
static void           *timer_id;
static int            kern_cp_time_oid[3], kern_cp_times_oid[3];
static int            smpActive = 0, smpCpus = 0, smpMaxCpus = 0;
static long           *cp_ptime = NULL, *cp_pload = NULL, cp_ptime_len = 0;

struct cpuEntry { 
	int32_t index;
	TAILQ_ENTRY(cpuEntry) link;
};

TAILQ_HEAD(cpuTable, cpuEntry);

static struct cpuTable cpuTable = TAILQ_HEAD_INITIALIZER(cpuTable);

int
cpuload_get(struct snmp_context __unused *ctx, struct snmp_value *val,
	u_int sub, u_int __unused vindex, enum snmp_op op)
{ 
	asn_subid_t which = val->var.subs[sub-1];
	int index=0;
	
	if (op == SNMP_OP_SET) 
		return SNMP_ERR_NOT_WRITEABLE;

	if (op != SNMP_OP_GET)
		return SNMP_ERR_GENERR;

	switch(which) { 
		case LEAF_cpuUser: 
			index=CP_USER;
			break;
		case LEAF_cpuNice: 
			index=CP_NICE;
			break;
		case LEAF_cpuSyst: 
			index=CP_SYS;
			break;
		case LEAF_cpuIntr: 
			index=CP_INTR;
			break;
		case LEAF_cpuIdle:
			index=CP_IDLE;
			break;
		default: 
			return SNMP_ERR_NOSUCHNAME;
	};

	val->v.uint32 = cp_load[index];
	return SNMP_ERR_NOERROR;
};

int
cpupercpu_get(struct snmp_context __unused *ctx, struct snmp_value *val,
	u_int sub, u_int __unused vindex, enum snmp_op op)
{ 
	int index=0;
	struct cpuEntry* cpuEntry;
	
	if (op == SNMP_OP_SET) 
		return SNMP_ERR_NOT_WRITEABLE;

	if (op != SNMP_OP_GET && op != SNMP_OP_GETNEXT) { 
		return SNMP_ERR_GENERR;
	};

	switch (op) { 
		case SNMP_OP_GETNEXT: 
			if((cpuEntry = NEXT_OBJECT_INT(&cpuTable, &val->var, 
				sub)) == NULL) 
				return SNMP_ERR_NOSUCHNAME;

			val->var.len = sub +1;
			val->var.subs[sub] = cpuEntry->index;
			break;
		case SNMP_OP_GET: 
			if ((cpuEntry = FIND_OBJECT_INT(&cpuTable, &val->var, 
				sub)) == NULL)
				return SNMP_ERR_NOSUCHNAME;
			break;
		default: 
printf("op is not get nor getnext\n");
			return SNMP_ERR_GENERR;
	};

	switch(val->var.subs[sub-1]) { 
		case LEAF_cpuUser: 
			index=CP_USER;
			break;
		case LEAF_cpuNice: 
			index=CP_NICE;
			break;
		case LEAF_cpuSyst: 
			index=CP_SYS;
			break;
		case LEAF_cpuIntr: 
			index=CP_INTR;
			break;
		case LEAF_cpuIdle:
			index=CP_IDLE;
			break;
		case LEAF_cpuIndex: 
			val->v.uint32 = cpuEntry->index;
			return SNMP_ERR_NOERROR;
		default: 
			return SNMP_ERR_NOSUCHNAME;
	};

	val->v.uint32 = cp_pload[(cpuEntry->index-1)*CPUSTATES+index];
	return SNMP_ERR_NOERROR;
};

int
cpusmp_get(struct snmp_context __unused *ctx, struct snmp_value *val,
	u_int sub, u_int __unused vindex, enum snmp_op op)
{ 
	asn_subid_t which = val->var.subs[sub-1];
	
	if (op == SNMP_OP_SET) 
		return SNMP_ERR_NOT_WRITEABLE;

	if (op != SNMP_OP_GET)
		return SNMP_ERR_GENERR;

	switch(which) { 
		case LEAF_cpuSmpActive: 
			val->v.uint32 = smpActive;
			break;
		case LEAF_cpuSmpCpus: 
			val->v.uint32 = smpCpus;
			break;
		case LEAF_cpuSmpMaxCpus: 
			val->v.uint32 = smpMaxCpus;
			break;
		default: 
			return SNMP_ERR_NOSUCHNAME;
	};

	return SNMP_ERR_NOERROR;
};

static void
cpuload_tick(void __unused *arg)
{ 
	long int cp_time1[CPUSTATES];
	long int delta = 0, j=0, i;
	size_t len = sizeof(cp_time1);
	long int cp_ptime1[smpMaxCpus*CPUSTATES];
	
	sysctl(kern_cp_time_oid, 2, cp_time1, &len, NULL, 0);

	for(j=0; j<CPUSTATES; j++) 
		delta+=cp_time1[j]-cp_time[j];

	for(j=0; j<CPUSTATES; j++) { 
		if(!delta) 
			cp_load[j]=0;
		else 
			cp_load[j]=(cp_time1[j]-cp_time[j])*100/delta;
		cp_time[j]=cp_time1[j];
	};

	len = sizeof(cp_ptime1);
	memset(cp_ptime1, 0, len);

	sysctl(kern_cp_times_oid, 2, cp_ptime1, &len, NULL, 0);

	for(i=0; i<smpCpus; i++) { 
		delta = 0;
		for(j=0; j<CPUSTATES; j++) 
			delta+=cp_ptime1[i*CPUSTATES+j]-cp_ptime[i*CPUSTATES+j];

		for(j=0; j<CPUSTATES; j++) {
			if(!delta) 
				cp_pload[i*CPUSTATES+j]=0;
			else 
				cp_pload[i*CPUSTATES+j]=
					(cp_ptime1[i*CPUSTATES+j]-
					cp_ptime[i*CPUSTATES+j])*100/delta;
			cp_ptime[i*CPUSTATES+j]=cp_ptime1[i*CPUSTATES+j];
		};
	};
		
	return ;
};

static int 
cpuload_init(struct lmodule* mod, int __unused argc, char __unused *argv[])
{ 
	size_t len = 3;
	int i;

	module = mod;

	if(sysctlnametomib("kern.cp_time", kern_cp_time_oid, &len) == -1) {
		syslog(LOG_ERR, "cpuload_init(): kern.cp_time: %s\n", 
			strerror(errno));
		return -1;
	};

	if(sysctlnametomib("kern.cp_times", kern_cp_times_oid, &len) == -1) {
		syslog(LOG_ERR, "cpuload_init(): kern.cp_times: %s\n", 
			strerror(errno));
		return -1;
	};

	len = sizeof(cp_time);

	if(sysctl(kern_cp_time_oid, 2, cp_time, &len, NULL, 0) ==-1){
		syslog(LOG_ERR, "cpuload_init(): unable to read kern.cp_time: "
			"%s\n", strerror(errno));
		return -1;
	};

	len = sizeof(smpActive);

	if(sysctlbyname("kern.smp.active", &smpActive, &len, NULL, 0) == -1) {
		syslog(LOG_ERR, "cpuload_init(): unable to read kern.smp.active"
			": %s\n", strerror(errno));
	};
	if(sysctlbyname("kern.smp.cpus", &smpCpus, &len, NULL, 0) == -1) {
		syslog(LOG_ERR, "cpuload_init(): unable to read kern.smp.cpus"
			": %s\n", strerror(errno));
	};
	if(sysctlbyname("kern.smp.maxcpus", &smpMaxCpus, &len, NULL, 0) == -1) {
		syslog(LOG_ERR, "cpuload_init(): unable to read kern.smp."
			"maxcpus: %s\n", strerror(errno));
	} else { 
		cp_ptime_len = sizeof(long int)*CPUSTATES*smpMaxCpus;
		cp_ptime = malloc(cp_ptime_len);
		if(!cp_ptime) { 
			syslog(LOG_ERR, "cpuload_init(): unable to allocate %i"
				" bytes to hold stats of %i CPUs\n", 
				cp_ptime_len, smpMaxCpus);
			return -1;
		};

		len = cp_ptime_len;

		if(sysctl(kern_cp_times_oid, 2, cp_ptime, &len, NULL, 0)==-1) {
			syslog(LOG_ERR, "cpuload_init(): unable to read "
				"kern.cp_times: %s\n", strerror(errno));
			return -1;
		};

		cp_pload = malloc(cp_ptime_len);
		if(!cp_pload) { 
			syslog(LOG_ERR, "cpuload_init(): unable to allocate %i"
				" bytes to hold load stats of %i CPUs\n", 
				cp_ptime_len, smpMaxCpus);
			return -1;
		};
	};

	for (i=0 ; i<smpCpus; i++) { 
		struct cpuEntry* cp=malloc(sizeof(struct cpuEntry));
		cp->index=i+1;
		INSERT_OBJECT_INT(cp, &cpuTable);
	};

	cpuload_tick(NULL);
		
	timer_id=timer_start_repeat(6000, 6000, cpuload_tick, NULL, module);

	return 0;
};

static int
cpuload_fini(void)
{ 
	if (timer_id != NULL) 
		timer_stop(timer_id);
	timer_id = NULL;

	if (cp_ptime != NULL) 
		free(cp_ptime);
	cp_ptime = NULL;

	if (cp_pload != NULL) 
		free(cp_pload);
	cp_pload = NULL;

	while(!TAILQ_EMPTY(&cpuTable)) { 
		struct cpuEntry* cpuEntry = TAILQ_FIRST(&cpuTable);
		TAILQ_REMOVE(&cpuTable, cpuEntry, link);
		free(cpuEntry);
	};

	return 0;
};
	
const struct snmp_module config = { 
	.comment = "This module implements CPU load MIB",
	.init    = cpuload_init,
	.fini    = cpuload_fini,
	.tree    = cpuload_ctree,
	.tree_size = cpuload_CTREE_SIZE,
};
	
